from __future__ import division

import time

from .module import Module


class Stepper(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Stepper', id, alias, robot)

        # Read
        self._position = None

        # Write
        self._target_position = None
        self._target_speed = None

    @property
    def position(self):
        """ Current position in mm """
        return self._position

    @property
    def is_moving(self):
        """ Is the stepper moving """
        # WARN: This function will not work properly while
        # this info is not fetched from the module itself
        return self._position != self._target_position

    def wait_until_idle(self):
        while self.is_moving:
            time.sleep(0.1)

    @property
    def target_position(self):
        """ Target position in mm. """
        return self._target_position

    @target_position.setter
    def target_position(self, new_pos):
        self._target_position = new_pos
        self._push_value('target_position', self._target_position)

    @property
    def target_speed(self):
        """ Speed in mm per seconds. """
        return self._target_speed

    @target_speed.setter
    def target_speed(self, new_speed):
        if new_speed != self._target_speed:
            self._target_speed = new_speed
            self._push_value('target_speed', self._target_speed)

    def home(self):
        self._push_value('home', 0)

    def stop(self):
        self._push_value('stop', 0)

    def _update(self, new_state):
        new_pos = new_state['position']

        if new_pos != self._position:
            self._pub_event('moved', self._position, new_pos)
            self._position = new_pos



from .generic_motor_controller import GenericMotorController
from .distance import Distance
from .l0_servo import L0Servo
from .encoder import Encoder
from .stepper import Stepper
from .l0_gpio import L0GPIO
from .l0_dc_motor import L0DCMotor
from .dxl import Dynamixel
from .button import Button
from .potard import Potard
from .servo import Servo
from .relay import Relay
from .led import Led

__all__ = [
    'name2mod',
    'GenericMotorController',
    'Dynamixel',
    'Distance',
    'Encoder',
    'Stepper',
    'Button',
    'L0GPIO',
    'L0Servo',
    'Potard',
    'Servo',
    'Relay',
    'Led',
    'L0DCMotor',
]

name2mod = {
    'generic_motor_controller': GenericMotorController,
    'dynamixel': Dynamixel,
    'distance': Distance,
    'l0_gpio': L0GPIO,
    'l0_servo': L0Servo,
    'l0_dc_motor': L0DCMotor,
    'encoder': Encoder,
    'stepper': Stepper,
    'button': Button,
    'potard': Potard,
    'servo': Servo,
    'relay': Relay,
    'led': Led,
}



from .module import Module, interact


class Led(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'LED', id, alias, robot)
        self.color = (0, 0, 0)

    @property
    def color(self):
        return self._value

    @color.setter
    def color(self, new_color):
        new_color = [int(min(max(c, 0), 255)) for c in new_color]

        if new_color != self._value:
            self._value = new_color
            self._push_value('rgb', new_color)

    def control(self):
        def change_color(red, green, blue):
            self.color = (red, green, blue)

        return interact(change_color,
                        red=(0, 255, 1),
                        green=(0, 255, 1),
                        blue=(0, 255, 1))



from __future__ import division

import time
import socket

from contextlib import closing
from threading import Thread, Event

from tornado.ioloop import IOLoop
from tornado.web import Application
from tornado.websocket import WebSocketHandler

from pyluos.tools import usb_gate, wifi_gate


class Publisher(WebSocketHandler):
    verbose = False

    gates = []
    pub = Event()

    def open(self):
        if self.verbose:
            print('WebSocket connection open.')

        self.publish_gates()
        self.pub_in_bg()

    def on_close(self):
        if self.verbose:
            print('WebSocket closed {}.'.format(self.close_reason))

    def publish_gates(self):
        self.ioloop.add_callback(self._real_pub)

    def _real_pub(self):
        if self.verbose:
            print('Send: {}'.format(Publisher.gates))

        self.write_message(Publisher.gates)

    def pub_in_bg(self):
        def _pub():
            while True:
                self.pub.wait()
                self.publish_gates()
                self.pub.clear()

        t = Thread(target=_pub)
        t.daemon = True
        t.start()

    def check_origin(self, origin):
        return True


def connect(host, port):
    loop = IOLoop()

    Publisher.verbose = False
    Publisher.ioloop = loop

    app = Application([
        (r'/', Publisher)
    ])

    app.listen(port)
    url = 'ws://{}:{}'.format(host, port)
    if Publisher.verbose:
        print('Fake robot serving on {}'.format(url))
    t = Thread(target=loop.start)
    t.daemon = True
    t.start()


def update(gates):
    Publisher.gates = gates
    Publisher.pub.set()


def free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.listen(1)
        port = s.getsockname()[1]
        return port


def main():
    usb_gates = {}
    redirect = {}
    usb2ws = {}

    wifi_gates = {}

    connect('127.0.0.1', 9090)

    while True:
        # Deal with USB gates
        found_usb_gates = usb_gate.discover()

        unplugged_gates = set(usb_gates) - set(found_usb_gates)
        for g in unplugged_gates:
            print('Gate "{}" unplugged.'.format(usb_gates[g]))

            print('--> Stop its redirection.')
            redirect.pop(g)
            usb2ws.pop(g).terminate()

        plugged_gates = set(found_usb_gates) - set(usb_gates)
        for g in plugged_gates:
            print('Gate "{}" plugged.'.format(found_usb_gates[g]))

            ws_port = free_port()
            print('--> Redirecting it to "ws://127.0.0.1:{}".'.format(ws_port))
            p = usb_gate.redirect_to_ws(found_usb_gates[g], ws_port)
            redirect[g] = ws_port
            usb2ws[g] = p

        usb_gates = found_usb_gates

        # Deal with WiFi gate
        found_wifi_gates = wifi_gate.discover()

        lost_gates = set(wifi_gates) - set(found_wifi_gates)
        for g in lost_gates:
            print('Gate "{}" lost.'.format(g))

        discovered_gates = set(found_wifi_gates) - set(wifi_gates)
        for g in discovered_gates:
            print('Gate "{}" discovered.'.format(g))

        wifi_gates = found_wifi_gates

        gates = {}
        for name in usb_gates.keys():
            gates[name] = {'host': '127.0.0.1', 'port': redirect[name]}

        for name, (host, port) in wifi_gates.items():
            gates[name] = {'host': host, 'port': port}

        update(gates)

        time.sleep(1.)


if __name__ == '__main__':
    main()



from __future__ import division

import json
import time
import serial as _serial

from threading import Event, Thread

from serial.tools.list_ports import comports

from . import IOHandler

black_list = [
    '/dev/cu.Bluetooth-Incoming-Port'
]

try:
    JSONDecodeError = json.decoder.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError


class Serial(IOHandler):
    poll_frequency = 100

    @classmethod
    def available_hosts(cls):
        return list(p.device for p in comports()
                    if p.device not in black_list)

    @classmethod
    def is_host_compatible(cls, host):
        return host in cls.available_hosts()

    def __init__(self, host, baudrate=57600):
        self._serial = _serial.Serial(host, baudrate)
        self._serial.flush()

        self._msg = None
        self._msg_ready = Event()
        self._running = True

        self._poll_loop = Thread(target=self._poll)
        self._poll_loop.daemon = True
        self._poll_loop.start()

    def is_ready(self):
        if self._serial.in_waiting == 0:
            return False

        try:
            self.read()
            return True
        except (UnicodeDecodeError, JSONDecodeError):
            return False

    def recv(self):
        self._msg_ready.wait()
        self._msg_ready.clear()
        return self._msg

    def write(self, data):
        self._serial.write(data + '\r'.encode())

    def close(self):
        self._running = False
        self._poll_loop.join()

        self._serial.close()

    def _poll(self):
        def extract_last_line(s):
            j = s.rfind(b'\n')

            if j == -1:
                return b'', s

            i = s[:j].rfind(b'\n') + 1
            return s[i:j], s[j + 1:]

        period = 1 / self.poll_frequency
        buff = b''

        while self._running:
            to_read = self._serial.in_waiting

            if to_read == 0:
                time.sleep(period)
                continue

            s = self._serial.read(to_read)
            line, buff = extract_last_line(buff + s)
            if len(line):
                self._msg = line
                self._msg_ready.set()



from .module import Module


class Distance(Module):
    possible_events = {'changed', 'filter_changed'}
    threshold = 10

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Distance', id, alias, robot)
        self._value = 0

    @property
    def distance(self):
        """ Distance in mm. """
        return self._value

    def _update(self, new_state):
        new_dist = new_state['distance']

        if new_dist != self._value:
            self._pub_event('changed', self._value, new_dist)

            if abs(new_dist - self._value) > self.threshold:
                self._pub_event('filter_changed',
                                self._value, new_dist)

            self._value = new_dist



from .module import Module


class Potard(Module):
    possible_events = {'moved'}

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Potard', id, alias, robot)
        self._value = 0

    @property
    def position(self):
        """ Position in degrees. """
        return self._value / 2048.0 - 1.0

    def _update(self, new_state):
        new_pos = new_state['position']

        if new_pos != self._value:
            self._value = new_pos
            self._pub_event('moved', self._value, self.position)



import json

from threading import Thread

from tornado.ioloop import IOLoop
from tornado.web import Application
from tornado.websocket import WebSocketHandler

from pyluos.io.serial_io import Serial


class SerialToWs(WebSocketHandler):
    def open(self):
        print('Connection opened.')
        self.serial = Serial(self.serial_port)
        Thread(target=self._check_msg).start()

    def on_close(self):
        if self.verbose:
            print('Connection closed.')
        self.serial.close()

    def on_message(self, message):
        if self.verbose:
            print('WS->Ser: {}'.format(message))
        self.serial.write(message.encode())

    def send(self, message):
        if self.verbose:
            print('Ser->WS: {}'.format(message))
        self.ioloop.add_callback(lambda: self.write_message(message))

    def _check_msg(self):
        while True:
            r = self.serial.recv()
            try:
                json.loads(r)
            except:
                print('Dump {}'.format(r))
                continue
            self.send(r)

    def check_origin(self, origin):
        return True


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-port', type=str, required=True)
    parser.add_argument('--ws-port', type=int, required=True)
    parser.add_argument('--verbose', action='store_true', default=False)
    args = parser.parse_args()

    loop = IOLoop()

    SerialToWs.serial_port = args.serial_port
    SerialToWs.verbose = args.verbose
    SerialToWs.ioloop = loop

    app = Application([
        (r'/', SerialToWs)
    ])

    app.listen(args.ws_port)
    loop.start()


if __name__ == '__main__':
    main()



from .module import Module


class Encoder(Module):
    possible_events = {'moved'}

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Encoder', id, alias, robot)
        self._value = 0

    @property
    def position(self):
        """ Position in degrees """
        return self._value

    def home(self):
        self._push_value('home', 0)

    def _update(self, new_state):
        new_pos = new_state['position']

        if new_pos != self._value:
            self._pub_event('moved', self._value, new_pos)
            self._value = new_pos



from .module import Module


class Button(Module):
    possible_events = {'changed', 'pressed', 'released'}

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Button', id, alias, robot)
        self._value = 'OFF'

    @property
    def state(self):
        return self._value

    def _update(self, new_state):
        new_state = 'ON' if new_state['state'] else 'OFF'

        if new_state != self._value:
            self._pub_event('changed', self._value, new_state)

            evt = 'pressed' if new_state == 'ON' else 'released'
            self._pub_event(evt, self._value, new_state)

            self._value = new_state



from collections import defaultdict, namedtuple

import logging

logger = logging.getLogger(__name__)

try:
    from ipywidget import interact
except ImportError:
    def interact(*args, **kwargs):
        logger.warning('You first have to install ipywidgets to use the control widgets.')
        return None


Event = namedtuple('Event', ('name', 'old_value', 'new_value'))


class Module(object):
    possible_events = set()

    def __init__(self,
                 type, id, alias,
                 robot):
        self.id = id
        self.type = type
        self.alias = alias
        self._delegate = robot
        self._value = None
        self._cb = defaultdict(list)

    def __repr__(self):
        return ('<{self.type} '
                'alias="{self.alias}" '
                'id={self.id} '
                '{state}>'.format(self=self,
                                  state=self._state_repr()))

    def _state_repr(self):
        return ('state={}'.format(self._value)
                if self._value is not None else
                '')

    def _update(self, new_state):
        pass

    def _push_value(self, key, new_val):
        self._delegate.update_cmd(self.alias, key, new_val)

    # Events cb handling

    def add_callback(self, event, cb):
        if event not in self.possible_events:
            raise ValueError('Unknown callback: {} (poss={})'.format(event, self.possible_events))

        self._cb[event].append(cb)

    def remove_callback(self, event, cb):
        self._cb[event].remove(cb)

    def _pub_event(self, trigger, old_value, new_value):
        event = Event(trigger, old_value, new_value)

        for cb in self._cb[trigger]:
            cb(event)



from __future__ import division

from .module import Module


class XL320(object):
    def __init__(self, name, delegate):
        self._name = name
        self._delegate = delegate
        self._target_pos = None
        self._moving_speed = None
        self._compliant = None

    @property
    def name(self):
        return self._name

    @property
    def position(self):
        return from_dxl_pos(self._position)

    @property
    def target_position(self):
        return self._target_pos

    @target_position.setter
    def target_position(self, new_pos):
        if new_pos != self._target_pos:
            self._delegate._push_value(self.name, to_dxl_pos(new_pos))
            self._target_pos = new_pos

    @property
    def moving_speed(self):
        return self._moving_speed

    @moving_speed.setter
    def moving_speed(self, speed):
        if speed != self._moving_speed:
            self._delegate._push_value(self.name.replace('m', 's'), to_dxl_speed(speed))
            self._moving_speed = speed

    @property
    def compliant(self):
        return self._compliant

    @compliant.setter
    def compliant(self, compliant):
        compliant = True if compliant else False
        if compliant != self._compliant:
            self._delegate._push_value(self.name.replace('m', 'c'), compliant)
            self._compliant = compliant


def to_dxl_pos(pos):
    pos = min(max(pos, -150.0), 149.9)
    return int((150.0 + pos) / 300.0 * 1024)


def from_dxl_pos(pos):
    pos = min(max(pos, 0), 1023)
    return (pos / 1024.0) * 300.0 - 150.0


def to_dxl_speed(speed):
    direction = 1024 if speed < 0 else 0
    speed_factor = 0.111
    max_value = 1023 * speed_factor * 6
    speed = min(max(speed, -max_value), max_value)
    return int(round(direction + abs(speed) / (6 * speed_factor), 0))


class Dynamixel(Module):
    MAX_MOTOR = 20

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Dynamixel', id, alias, robot)
        self._setup = False
        self.motors = []

    def _update(self, new_state):
        if not self._setup:
            motors = ['m{}'.format(i) for i in range(self.MAX_MOTOR)]
            motors = [m for m in motors if m in new_state]
            for m in motors:
                setattr(self, m, XL320(m, self))
                self.motors.append(getattr(self, m))

            self._setup = True

        for m in self.motors:
            if m.name in new_state:
                m._position = new_state[m.name]



import sys
import json
import time
import random
import threading

from copy import deepcopy
from datetime import datetime
from collections import defaultdict

from .io import io_from_host
from .modules import name2mod
from .metrics import Publisher


def run_from_unittest():
    return 'unittest' in sys.modules


use_topographe = True
try:
    import zmq
except ImportError:
    use_topographe = False


class Robot(object):
    _heartbeat_timeout = 5  # in sec.
    _max_alias_length = 15

    def __init__(self, host,
                 verbose=True, test_mode=False,
                 *args, **kwargs):
        self._io = io_from_host(host=host,
                                *args, **kwargs)

        self._verbose = verbose

        self._log('Connected to "{}".'.format(host))

        self._send_lock = threading.Lock()
        self._cmd_lock = threading.Lock()

        # We force a first poll to setup our model.
        self._setup()
        self._log('Robot setup.')

        self._last_update = time.time()
        self._running = True

        # Setup both poll/push synchronization loops.
        self._poll_bg = threading.Thread(target=self._poll_and_up)
        self._poll_bg.daemon = True
        self._poll_bg.start()

        if not (test_mode or run_from_unittest()):
            self._metrics_pub = Publisher(robot=self)
            self._metrics_pub.start()

            if use_topographe:
                c = zmq.Context()
                s = c.socket(zmq.PUB)
                s.connect('tcp://127.0.0.1:33000')
                self._s = s

    @property
    def state(self):
        return {
            'gate': self.name,
            'timestamp': datetime.now(),
            'types': ','.join([mod.type for mod in self.modules]),
            'modules': ','.join([mod.alias for mod in self.modules])
        }

    @property
    def name(self):
        return self._name

    @property
    def alive(self):
        dt = time.time() - self._last_update
        return self._running and dt < self._heartbeat_timeout

    def close(self):
        self._running = False
        self._poll_bg.join()
        self._io.close()

    def _setup(self):
        self._log('Sending detection signal.')
        self._send({'detection': {}})

        self._log('Waiting for first state...')
        while not self._io.is_ready():
            self._send({'detection': {}})

        state = self._poll_once()

        gate = next(g for g in state['modules']
                    if g['type'] == 'gate')
        self._name = gate['alias']

        modules = [mod for mod in state['modules']
                   if mod['type'] in name2mod.keys()]

        self._cmd = defaultdict(lambda: defaultdict(lambda: None))

        self.modules = [
            name2mod[mod['type']](id=mod['id'],
                                  alias=mod['alias'],
                                  robot=self)
            for mod in modules
        ]

        for mod in self.modules:
            setattr(self, mod.alias, mod)

        # We push our current state to make sure that
        # both our model and the hardware are synced.
        self._push_once()

    # Poll state from hardware.
    def _poll_once(self):
        self._state = self._io.read()
        self._state['timestamp'] = time.time()
        return self._state

    def _poll_and_up(self):
        while self._running:
            state = self._poll_once()
            self._update(state)
            self._push_once()

            self._broadcast(state)

    # Update our model with the new state.
    def _update(self, new_state):
        mod_need_update = [mod for mod in new_state['modules']
                           if hasattr(self, mod['alias']) and
                           set(mod.keys()) != {'type', 'id', 'alias'}]

        for mod in mod_need_update:
            getattr(self, mod['alias'])._update(mod)

        self._last_update = time.time()

    # Push update from our model to the hardware
    def _split_messages(self, cmd):
        alias = random.choice(list(cmd.keys()))
        mod = getattr(self, alias)

        cmd = deepcopy(cmd)

        if mod.type == 'Servo':
            to_send = {alias: cmd[alias]}
            del cmd[alias]
            rest = cmd

        else:
            servos = [
                mod for mod in cmd.keys()
                if getattr(self, mod).type == 'Servo'
            ]
            others = [
                mod for mod in cmd.keys()
                if getattr(self, mod).type != 'Servo'
            ]
            to_send = {
                mod: cmd[mod]
                for mod in others
            }
            rest = {
                mod: cmd[mod]
                for mod in servos
            }

        return to_send, rest

    def update_cmd(self, alias, key, val):
        with self._cmd_lock:
            self._cmd[alias][key] = val

    def _push_once(self):
        with self._cmd_lock:
            if not self._cmd:
                return

            to_send, remainings = self._split_messages(self._cmd)

            self._send({
                'modules': to_send
            })

            self._cmd = defaultdict(lambda: defaultdict(lambda: None))
            for mod, vals in remainings.items():
                for key, val in vals.items():
                    self._cmd[mod][key] = val

    def _send(self, msg):
        with self._send_lock:
            self._io.send(msg)

    def _broadcast(self, state):
        if not hasattr(self, '_s'):
            return

        for mod in state['modules']:
            if ((mod['type'] == 'servo') and
               (hasattr(self, mod['alias']))):
                servo = getattr(self, mod['alias'])
                mod['position'] = servo.target_position

        msg = '{} {}'.format(self.name, json.dumps(state))
        self._s.send_string(msg)

    def rename_module(self, old, new):
        if not hasattr(self, old):
            raise ValueError('No module named {}!'.format(old))

        if len(new) > self._max_alias_length:
            raise ValueError('Alias length should be less than {}'.format(self._max_alias_length))

        self._send({'modules': {old: {'set_alias': new}}})

        mod = getattr(self, old)
        mod.alias = new
        setattr(self, new, mod)
        delattr(self, old)

    def _log(self, msg):
        if self._verbose:
            print(msg)



from .module import Module


class Relay(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Relay', id, alias, robot)
        self.off()

    @property
    def state(self):
        return 'on' if self._value == 1 else 'off'

    def on(self):
        self._set(1)

    def off(self):
        self._set(0)

    def _set(self, new_val):
        if new_val != self._value:
            self._value = new_val
            self._push_value('state', new_val)



from __future__ import division

from .module import Module


class DCMotor(object):
    def __init__(self, name, delegate):
        self._name = name
        self._delegate = delegate
        self._speed = None

    @property
    def name(self):
        return self._name

    @property
    def speed(self):
        self._speed

    @speed.setter
    def speed(self, s):
        s = min(max(s, -1.0), 1.0)

        if s != self._speed:
            self._speed = s
            field = self.name.replace('m', 's')
            self._delegate._push_value(field, self._speed)


class L0DCMotor(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'L0DCMotor', id, alias, robot)

        self.m1 = DCMotor('m1', self)
        self.m2 = DCMotor('m2', self)



from .module import Module, interact
from .gpio import Pwm


class L0Servo(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'L0Servo', id, alias, robot)

        self.pwm_1 = Pwm('p1', self, max=180.0)
        self.pwm_2 = Pwm('p2', self, max=180.0)
        self.pwm_3 = Pwm('p3', self, max=180.0)
        self.pwm_4 = Pwm('p4', self, max=180.0)

    def control(self):
        def change_pin(p1, p2, p3, p4):
            self.p1._push(p1)
            self.p2._push(p2)
            self.p3._push(p3)
            self.p4._push(p4)

        return interact(change_pin,
                        p1=lambda: self.pwm_1.duty_cycle,
                        p2=lambda: self.pwm_2.duty_cycle,
                        p3=lambda: self.pwm_3.duty_cycle,
                        p4=lambda: self.pwm_4.duty_cycle)



import time

from contextlib import closing

from zeroconf import ServiceBrowser, Zeroconf


class MyListener(object):
    wifi_gates = set()

    def remove_service(self, zeroconf, type, name):
        self.wifi_gates.remove(name)

    def add_service(self, zeroconf, type, name):
        self.wifi_gates.add(name)


def discover():
    with closing(Zeroconf()) as zeroconf:
        listener = MyListener()
        listener.wifi_gates.clear()

        browser = ServiceBrowser(zeroconf, "_jsongate._tcp.local.", listener)

        time.sleep(1.0)

        gates = [g.replace('._jsongate._tcp', '')[:-1]
                 for g in listener.wifi_gates]

        return {
            host.replace('.local', ''): (host, 9342)
            for host in gates
        }


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('cmd', choices=('discover', ))
    args = parser.parse_args()

    if args.cmd == 'discover':
        while True:
            print(discover())


if __name__ == '__main__':
    main()



from .module import Module, interact


class GenericMotorController(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'generic_motor_controller', id, alias, robot)
        self._power = None

    @property
    def power_ratio(self):
        return self._power

    @power_ratio.setter
    def power_ratio(self, new_power):
        new_power = min(max(new_power, -100.0), 100.0)

        if new_power != self._power:
            self._power = new_power
            self._push_value('power_ratio', new_power)

    def control(self):
        def set_power(power):
            self.power_ratio = power

        return interact(set_power, power=(-100.0, 100.0, 1.0))



import json


class IOHandler(object):
    @classmethod
    def is_host_compatible(cls, host):
        return False

    def __init__(self, host):
        raise NotImplementedError

    def is_ready(self):
        raise NotImplementedError

    def read(self):
        data = self.recv()
        return self.loads(data)

    def recv(self):
        raise NotImplementedError

    def send(self, msg):
        self.write(self.dumps(msg))

    def write(self, data):
        raise NotImplementedError

    def loads(self, data):
        if type(data) == bytes:
            data = data.decode()
        return json.loads(data)

    def dumps(self, msg):
        return json.dumps(msg).encode()


from .ws import Ws
from .serial_io import Serial


def io_from_host(host, *args, **kwargs):
    for cls in [Ws, Serial]:
        if cls.is_host_compatible(host):
            return cls(host=host, *args, **kwargs)
    raise ValueError('No corresponding IO found.')



from __future__ import division

from .module import Module, interact


class Servo(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Servo', id, alias, robot)

    @property
    def target_position(self):
        return self._value

    @target_position.setter
    def target_position(self, new_pos):
        if new_pos != self._value:
            self._value = new_pos
            self._push_value('target_position', new_pos)

    @property
    def target_speed(self):
        return self.target_position / 180 * 200 - 100

    @target_speed.setter
    def target_speed(self, new_speed):
        pos = (new_speed + 100) / 200 * 180
        self.target_position = pos

    def control(self):
        def move(position):
            self.target_position = position

        return interact(move, position=(0, 180, 1))



import json
import time
import logging

import requests

from threading import Thread
from datetime import datetime

logger = logging.getLogger(__name__)


def json_handler(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()


class Publisher(object):
    def __init__(self, robot,
                 url='http://teahupoo.ddns.net/post/robot-state.json',
                 push_period=60):

        def pub_loop():
            while True:
                time.sleep(push_period)

                if not robot.alive:
                    continue

                try:
                    requests.post(
                        url=url,
                        data=json.dumps(robot.state,
                                        default=json_handler),
                        headers={
                            'content-type': 'application/json',
                        }
                    )
                except Exception as e:
                    msg = 'Could not post robot state: {}'.format(str(e))
                    logger.info(msg)

        self._pub = Thread(target=pub_loop)
        self._pub.daemon = True

    def start(self):
        self._pub.start()



from __future__ import division

from threading import Thread
from time import time, sleep
from math import sin, pi


class Sinus(object):
    update_frequency = 25.0

    def __init__(self, motor, frequency, amplitude, offset, phase):
        self.motor = motor

        self.frequency = frequency
        self.amplitude = amplitude
        self.offset = offset
        self.phase = phase

        self._running = False
        self._t = None

    def start(self):
        if self._t is not None:
            raise EnvironmentError('Sinus already running!')

        self._running = True
        self._t = Thread(target=self._run)
        self._t.start()

    def stop(self):
        self._running = False
        if self._t is not None:
            self._t.join()
            self._t = None

    def _run(self):
        t0 = time()

        while self._running:
            t = time() - t0
            pos = self.amplitude * sin(2 * pi * self.frequency * t + (self.phase * pi / 180)) + self.offset

            self.motor.target_position = pos
            sleep(1 / self.update_frequency)



import socket
import websocket

from . import IOHandler


def resolve_hostname(hostname, port):
    # We do our own mDNS resolution
    # to enforce we only search for IPV4 address
    # and avoid a 5s timeout in the websocket on the ESP
    # See https://github.com/esp8266/Arduino/issues/2110

    addrinfo = socket.getaddrinfo(hostname, port,
                                  socket.AF_INET, 0,
                                  socket.SOL_TCP)
    addr = addrinfo[0][4][0]
    return addr


class Ws(IOHandler):
    @classmethod
    def is_host_compatible(cls, host):
        try:
            socket.inet_pton(socket.AF_INET, host)
            return True
        except socket.error:
            return host.endswith('.local')

    def __init__(self, host, port=9342):
        host = resolve_hostname(host, port)
        url = 'ws://{}:{}'.format(host, port)

        self._ws = websocket.create_connection(url)

    def is_ready(self):
        return True

    def recv(self):
        return self._ws.recv()

    def write(self, data):
        self._ws.send(data)

    def close(self):
        self._ws.close()



#!/usr/bin/env python

import imp

from setuptools import setup, find_packages

version = imp.load_source('pyluos.version', 'pyluos/version.py')


setup(name='pyluos',
      version=version.version,
      packages=find_packages(),
      install_requires=['future',
                        'websocket-client',
                        'pyserial>3',
                        'requests',
                        'tornado',
                        'zeroconf',
                        ],
      extras_require={
          'tests': [],
          'topographe': ['pyzmq'],
          'jupyter-integration': ['ipywidgets'],
      },
      entry_points={
          'console_scripts': [
              'pyluos-wifi-gate = pyluos.tools.wifi_gate:main',
              'pyluos-usb-gate = pyluos.tools.usb_gate:main',
              'pyluos-usb2ws = pyluos.tools.usb2ws:main',
              'pyluos-scratch-broker = pyluos.tools.scratch_broker:main',
          ],
      },
      )



class DigitalInputPin(object):
    def __init__(self, alias, default=False):
        self.alias = alias
        self._val = default

    def is_high(self):
        return self._val

    def is_low(self):
        return not self.is_high()

    def _update(self, val):
        self._val = (val == 1)

    def __repr__(self):
        return '<"{}" Input state="{}">'.format(
            self.alias,
            'high' if self.is_high() else 'low'
        )


class AnalogInputPin(object):
    def __init__(self, alias, default=0):
        self.alias = alias
        self._val = default

    def read(self):
        return self._val

    def _update(self, val):
        self._val = val

    def __repr__(self):
        return '<"{}" Input value="{}">'.format(
            self.alias,
            self.read()
        )


class DigitalOutputPin(object):
    def __init__(self, alias, delegate, default=False):
        self.alias = alias
        self._delegate = delegate
        self._val = default

    def is_high(self):
        return self._val

    def set_high(self):
        self._push(True)

    def is_low(self):
        return not self.is_high()

    def set_low(self):
        self._push(False)

    def toggle(self):
        if self.is_high():
            self.set_low()
        else:
            self.set_high()

    def _push(self, val):
        if self._val != val:
            self._val = val
            self._delegate._push_value(self.alias, self._val)

    def __repr__(self):
        return '<"{}" Output state="{}">'.format(
            self.alias,
            'high' if self.is_high() else 'low'
        )


class Pwm(object):
    def __init__(self, alias, delegate, default=0.0, min=0.0, max=1.0):
        self.alias = alias
        self._delegate = delegate
        self._val = default
        self._min, self._max = min, max

    @property
    def duty_cycle(self):
        return self._val

    @duty_cycle.setter
    def duty_cycle(self, duty):
        self._push(min(max(duty, self._min), self._max))

    def _push(self, val):
        if self._val != val:
            self._val = val
            self._delegate._push_value(self.alias, self._val)



from .module import Module, interact
from .gpio import AnalogInputPin, DigitalInputPin, DigitalOutputPin, Pwm


class L0GPIO(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'L0GPIO', id, alias, robot)

        self.pwm_1 = Pwm('p1', self)
        self.pwm_2 = Pwm('p2', self)
        self.digital_3 = DigitalOutputPin('p3', self)
        self.digital_4 = DigitalOutputPin('p4', self)

        self.digital_5 = DigitalInputPin('p5')
        self.digital_6 = DigitalInputPin('p6')
        self.analog_7 = AnalogInputPin('p7')
        self.analog_8 = AnalogInputPin('p8')
        self.analog_9 = AnalogInputPin('p9')

    def _update(self, new_state):
        self.digital_5._update(new_state['p5'])
        self.digital_6._update(new_state['p6'])
        self.analog_7._update(new_state['p7'])
        self.analog_8._update(new_state['p8'])
        self.analog_9._update(new_state['p9'])

    def control(self):
        def change_pin(p1, p2, p3, p4):
            self.p1._push(p1)
            self.p2._push(p2)
            self.p3._push(p3)
            self.p4._push(p4)

        return interact(change_pin,
                        p1=lambda: self.pwm_1.duty_cycle,
                        p2=lambda: self.pwm_2.duty_cycle,
                        p3=self.digital_3.is_high(),
                        p4=self.digital_4.is_high())
